// src/lib/create-params.svelte.ts
import { tick } from "svelte";

// src/lib/adapters/dom.ts
function dom(options = {}) {
  const { windowObj = window, replace = false } = options;
  const replaceState = windowObj.history.replaceState.bind(windowObj.history);
  const pushState = windowObj.history.pushState.bind(windowObj.history);
  const update = replace ? replaceState : pushState;
  return {
    isBrowser: () => typeof window !== "undefined",
    browser: {
      read: () => windowObj.location,
      save: (search, hash) => update(null, "", `${search}${hash}`)
    },
    server: {
      save: () => {
      }
    }
  };
}

// src/lib/utils.ts
import { parse } from "valibot";
function fromURL({ search }) {
  const params = new URLSearchParams(search);
  return Object.fromEntries(params.entries());
}
function parseObject(schemas, input) {
  const clone = {};
  const keys = /* @__PURE__ */ new Set([...Object.keys(input), ...Object.keys(schemas)]);
  for (const key of keys) {
    const value = input[key];
    clone[key] = key in schemas && schemas[key] ? parseValue(key, schemas[key], value) : (
      /** Value wasn't defined in the schema, pass through as-is */
      value
    );
  }
  return clone;
}
function parseValue(key, schema, value) {
  if (typeof schema === "function") return schema(value);
  if (isZodSchema(schema)) return schema.parse(value);
  if (isValibotSchema(schema)) return parse(schema, value);
  throw new Error(
    `Unknown schema type (${typeof schema}) for field: "${key}" (value: ${value})`
  );
}
function isZodSchema(obj) {
  return typeof obj === "object" && obj && "parse" in obj && "safeParse" in obj;
}
function isValibotSchema(obj) {
  return typeof obj === "object" && obj && "async" in obj && "_parse" in obj;
}
function parseQueryParams(params, schemas) {
  if (typeof schemas === "function") return schemas(params);
  if (isZodSchema(schemas)) return schemas.parse(params);
  if (isValibotSchema(schemas)) return parse(schemas, params);
  return parseObject(schemas, params);
}
function mapValues(object, mapFn) {
  return Object.fromEntries(
    Object.entries(object).map(([key, value]) => [key, mapFn(value)])
  );
}
function debounce(func, delay) {
  if (delay === 0) return func;
  let timeoutId;
  return (...args) => {
    clearTimeout(timeoutId);
    timeoutId = setTimeout(() => {
      func(...args);
    }, delay);
  };
}

// src/lib/create-params.svelte.ts
function createUseQueryParams(validators, options = {}) {
  const {
    debounce: delay = 0,
    windowObj = typeof window === "undefined" ? void 0 : window,
    adapter = dom({ windowObj }),
    serialise = (value) => typeof value === "string" ? value : JSON.stringify(value)
  } = options;
  let raw = $state({});
  const query = $derived(parseQueryParams(raw, validators));
  const search = $derived(`?${new URLSearchParams(query)}`);
  function readFromBrowser() {
    raw = fromURL(adapter.browser.read());
  }
  const persistToBrowser = debounce((search2, hash) => {
    return tick().then(() => adapter.browser.save(search2, hash));
  }, delay);
  function persistParams(hash) {
    adapter.isBrowser() ? (
      // By the time peristParams is called in the browser, a hash may have changed due to reactivity
      persistToBrowser(search, adapter.browser.read().hash)
    ) : (
      // We don't have this problem on the server
      adapter.server.save(search, hash)
    );
  }
  let unsubscribe = () => {
  };
  if (windowObj) {
    const replaceState = windowObj.history.replaceState.bind(windowObj.history);
    const pushState = windowObj.history.pushState.bind(windowObj.history);
    windowObj.history.replaceState = (...args) => {
      replaceState(...args);
      readFromBrowser();
    };
    windowObj.history.pushState = (...args) => {
      pushState(...args);
      readFromBrowser();
    };
    windowObj.addEventListener("popstate", readFromBrowser);
    unsubscribe = () => {
      if (windowObj) {
        windowObj.removeEventListener("popstate", readFromBrowser);
        windowObj.history.pushState = pushState;
        windowObj.history.replaceState = replaceState;
      }
    };
  }
  return function useQueryParams(url) {
    raw = fromURL({ search: url.search });
    const params = {};
    for (const key of Object.keys(validators)) {
      Object.defineProperty(params, key, {
        enumerable: true,
        configurable: true,
        get() {
          return query[key];
        },
        set(newValue) {
          raw = { ...raw, [key]: serialise(newValue) };
          persistParams(url.hash);
        }
      });
    }
    return [
      params,
      {
        get raw() {
          return raw;
        },
        get query() {
          return query;
        },
        get search() {
          return search;
        },
        get all() {
          return { ...raw, ...query };
        },
        keys() {
          return Object.keys(validators);
        },
        entries() {
          return Object.entries(validators).map(([key]) => [key, query[key]]);
        },
        set(params2) {
          raw = mapValues(params2, serialise);
          persistParams(url.hash);
        },
        update(params2) {
          raw = { ...raw, ...mapValues(params2, serialise) };
          persistParams(url.hash);
        },
        remove(...params2) {
          raw = Object.fromEntries(
            Object.entries(raw).filter(([key]) => !params2.includes(key))
          );
        },
        unsubscribe() {
          return unsubscribe();
        }
      }
    ];
  };
}
export {
  createUseQueryParams
};
