import { BaseIssue, BaseSchema, InferOutput } from 'valibot';
import { z } from 'zod';

interface Adapter {
    /**
     * @returns Whether we are in the browser
     */
    isBrowser(): boolean;
    browser: BrowserAdapter;
    server: ServerAdapter;
}
interface BaseAdapter {
    /**
     * A function to update the URL.
     *
     * Note: The first param includes the `?` prefix and the second param
     * includes the `#` if there is a hash
     */
    save: (search: string, hash: string) => void;
}
interface BrowserAdapter extends BaseAdapter {
    /**
     * A function to retrieve the URL when in the browser
     */
    read: () => URLLike;
}
interface ServerAdapter extends BaseAdapter {
}

type ValibotValidator<TInput = any, TOutput = TInput, TIssue extends BaseIssue<any> = any> = BaseSchema<TInput, TOutput, TIssue>;
type FunctionValidator<TOut extends object = any> = (value?: unknown) => TOut;
type FunctionValueValidator<TOut = any> = (value?: string) => TOut;
type ZodValidator = z.ZodType;
type ValueValidator = FunctionValueValidator | ZodValidator | ValibotValidator;
type Validator = FunctionValidator | ZodValidator | ValibotValidator;
type inferFromValidator<TValidator extends Validator> = TValidator extends ZodValidator ? z.infer<TValidator> : TValidator extends ValibotValidator ? InferOutput<TValidator> : TValidator extends FunctionValidator ? ReturnType<TValidator> : TValidator extends FunctionValueValidator ? ReturnType<TValidator> : never;
type QuerySchema = Validator | Record<string, ValueValidator>;
type Empty = Record<string, never>;
type inferShape<TShape extends QuerySchema> = TShape extends Validator ? inferFromValidator<TShape> : TShape extends Empty ? Empty : TShape extends Record<string, Validator> ? {
    [K in keyof TShape]: inferFromValidator<TShape[K]>;
} & {} : never;
/**
 * @param search Includes the `?` prefix
 */
type QueryUpdater = (search: string, hash: string) => void;
type Serializer = (value: unknown) => string;
type URLLike = URL | Location;
type WindowLike = Pick<typeof window, "location" | "history" | "addEventListener" | "removeEventListener">;
interface QueryParamsOptions {
    /**
     * Provide a custom implementation of {@linkcode window}. It requires
     * {@linkcode window.location}, {@linkcode window.history},
     * {@linkcode window.addEventListener} and {@linkcode window.removeEventListener}
     *
     * @default window
     */
    windowObj?: WindowLike;
    /**
     * Add a delay (in ms) before updating the browser URL. This is useful in
     * situations where URL updates happen frequently, e.g., on every keystroke.
     *
     * Note this does not affect the query params rune - this will always be
     * updated optimistically.
     *
     * @default 0
     */
    debounce?: number;
    /**
     * Control how query params are serialised to the browser query params
     *
     * **Note**: this is NOT for encoding values into URI components - it is
     * for serialising values into strings, which will then be encoded
     * internally.
     */
    serialise?: Serializer;
    adapter?: Adapter;
}
type QueryHelpers<TShape extends Record<string, unknown>> = {
    /**
     * The raw query params, extracted from {@linkcode windowObj.location.href}
     *
     * Note: this may include query params not defined in your schema. Values will
     * not have been parsed even if you have specified so in your validators.
     */
    readonly raw: Record<string, string>;
    /**
     * Similar to {@linkcode raw}, but any params specified in your validators
     * will have been parsed - all other values are passed through as-is.
     *
     * Note: this may include query params not defined in your schema.
     */
    readonly all: Record<string, string> & TShape;
    /**
     * The query string, generated from the {@linkcode query} which may contain
     * query params not defined in your schema. This always starts with a `?`.
     */
    readonly search: string;
    /** Replace _ALL_ query params, triggering a reactive and browser update */
    set(params: TShape): void;
    /** Update a subset of the query params, triggering a reactive and browser update */
    update(params: Partial<TShape>): void;
    /** Remove query params. Note that this may cause your validation check to fail */
    remove(...params: (keyof TShape)[]): void;
    /** Manually unset unregister all event listeners */
    unsubscribe(): void;
    /** Return the query keys. Unlike {@linkcode Object.keys}, this is type-safe */
    keys(): Array<keyof TShape>;
    entries(): Array<[keyof TShape, TShape[keyof TShape]]>;
};
type QueryHook<TShape extends Record<string, unknown>> = [
    TShape,
    QueryHelpers<TShape>
];
type UseQueryHook<TShape extends Record<string, unknown>> = 
/**
 * @param url The current URL
 */
(url: URLLike) => QueryHook<TShape>;

export type { Adapter as A, BrowserAdapter as B, FunctionValidator as F, QuerySchema as Q, Serializer as S, UseQueryHook as U, ValibotValidator as V, WindowLike as W, ZodValidator as Z, QueryParamsOptions as a, QueryHelpers as b, QueryHook as c, QueryUpdater as d, Validator as e, inferFromValidator as f, URLLike as g, ServerAdapter as h, inferShape as i, BaseAdapter as j };
